#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "clca.h"

int InterpretFlags(
/* Uses args to set up variables used to generate lca */
char *Args[],         /* Command line / param file args */
int Num,          /* Number of args */
int *Background,          /* Screen background, as determined by -b */
int *Palette,         /* Screen palette, as determined by -p */
int *Width,   /* Number of cells, already set to default; changed only if -w
                         argument is given */
int Rules[],          /* Rules used to generate new values of cells */
int *Gens,        /* Number of generations to print */
int *MutPeriod,       /* Number of generations between mutations */
int *Scroll,          /* Set to 1 if -s argument is found */
int *VidMode,		/* Video mode */
int *LineNap,		/* How long to nap between lines */
int *ScreenNap,		/* How long to nap between screens */
int Cells[],          /* Unpacked cells array */
ParamList *Params        /* Params from file given with -a arg */
)
{
    int I;
    int Error = 0,        /* Whether a bad value was given in any arg */
    BadValArg;        /* Whether a bad value was given with last arg */

    /* Remove - or / from in front of arguments */
    RemoveFlagSpecs(Args,Num,FLAG_SPECS);
    for (I = 0; I < Num; I++) {       /* For each arg given */
        BadValArg = 0;
        switch (*Args[I]) {
            case 'a':   BadValArg = !GetParams(Args[I] + 1,Params);
                        break;
            case 'b':   *Background = atoi(Args[I] + 1);
                        BadValArg = !(0 <= *Background && *Background <= 15);
                        break;
            case 'g':   *Gens = atoi(Args[I] + 1);
                        BadValArg = *Gens < 1;
                        break;
            case 'i':   BadValArg = !CharToInt(Cells,Args[I] + 1,MAX_CELLS,3);
                        break;
	    case 'l':	*LineNap = atoi(Args[I] + 1);
                        BadValArg = *LineNap < 0;
			break;
            case 'm':   *MutPeriod = atoi(Args[I] + 1);
                        BadValArg = *MutPeriod < 0;
                        break;
	    case 'n':	*ScreenNap = atoi(Args[I] + 1);
                        BadValArg = *ScreenNap < 0;
			break;
            case 'p':   *Palette = atoi(Args[I] + 1);
                        BadValArg = !(*Palette == 0 || *Palette == 1);
                        break;
            case 'r':   BadValArg = !CharToInt(Rules,Args[I] + 1,10,3);
                        break;
            case 's':   *Scroll = 1;
                        break;
	    case 'v':	*VidMode = Args[I][1];
			break;
            case 'w':   *Width = (atoi(Args[I] + 1) + 3) & 0xfffc;
                        BadValArg = !(3 < *Width && *Width <= MAX_CELLS);
                        break;
            default:    fputs("Bad command line argument: -",stdout);
                        puts(Args[I]);
                        Error = 1;
        }
        if (BadValArg) {
            Error = 1;
            fputs("Bad value given in -",stdout);
            putchar(*Args[I]);
            puts(" argument.");
        }
    }
    return(!Error);
}

int GetParams(char *FileName,ParamList *Params)
/* Reads lines from FileName and parses them into arrays of args as though
    they had been given on the command line; puts them in ParamList */
{
    FILE *InFile;         /* Input file */
    char InString[512];       /* String read from file */

    if (Params -> Next != NULL) {  /* attempt to insert more args in list */
        puts("Error - linked parameter files not allowed.");
        return 0;
    }
    InFile = fopen(FileName,"r");
    if (InFile == NULL)
        return 0;
    fgets(InString,512,InFile);       /* Read first string */
    while (!feof(InFile)) {
        if (!GetArgs(InString,Params -> Args,&(Params -> NumArgs),8)) {
            puts("Too many arguments in param file.");
            return 0;
        }
        /* Add another node to list */
        Params -> Next = MallocT(sizeof(ParamList));
        Params = Params -> Next;
        fgets(InString,512,InFile);       /* Read next string */
    }
    Params -> Next = NULL;        /* Terminate list */
    return 1;
}

int GetArgs(char *S,char *Args[],int *NumArgs,int MaxArgs)
/* Parse string S into and array of arguments, as though they were given on
    the command line */
{
    int I = 0;
    char *CurArg;         /* Current arg */

    CurArg = strtok(S," ");       /* Get first arg */
    while (CurArg != NULL) {
        if (I >= MaxArgs)
            return 0;
        Args[I++] = strdup(CurArg); /* Allocate space for arg and copy it */
        CurArg = strtok(NULL," ");        /* Get next arg */
    }
    *NumArgs = I;
    return 1;
}

void *MallocT(unsigned Size)
/* Calls malloc() to allocate a block of memory of size Size and checks
    whether there was enough memory. Exits if not, otherwise returns a
    pointer to the block */
{
void *Temp;

    Temp = (void *) malloc(Size);
    if (Temp == NULL) {
        puts("Error: Not enough memory.");
        exit(2);
    }
    return(Temp);
}

int CharToInt(
/* Convert a string of char representations of integer values, optionally
    separated by commas, to numeric integers and store them in an array */
int *Ints,        /* Result */
char *Chars,       /* String of digit chars representing integers */
int NumInts,          /* Number of ints needed */
int HighestInt)       /* Highest value */
{
    int I,J,Result;

    for (I = J = 0; Chars[I] != '\0' && J < NumInts; I++) {
        if (Chars[I] != ',') {
            Result = Chars[I] - '0';
            if (0 <= Result && Result <= HighestInt)
                Ints[J++] = Result;
            else
                return 0;
        }
    }
    for (; J < NumInts; J++) /* If not enough rules given, set the rest to 0 */
        Ints[J] = 0;
    return 1;
}

void RemoveFlagSpecs(char *Args[],int Num,char FlagSpecs[])
/* Removes any leading char that is in FlagSpecs from command line arguments
by incrementing the argument pointer */
{
    for (; Num; Num--)
        if (strchr(FlagSpecs,*(Args[Num - 1])) != NULL)
            Args[Num - 1]++;
}

int TranslateParams(ParamList *Params,EnvListType *EnvList)
/* Translates the arg arrays in Params, which are in the form of "argv",
    to a list of structures each of which specifies an lca environment. */
{
    int Cells[MAX_CELLS];         /* Cells array returned by InterpretFlags */
    EnvListType *CurEnv;                 /* Current env struct being set up */
    int Width;                      /* Width as returned by InterpretFlags */

    CurEnv = EnvList;         /* Start with EnvList */
    while (Params -> Next != NULL) {
        CurEnv -> Background = DEF_BACK;  /* Color of bckgnd, set to default */
        CurEnv -> Palette = DEF_PAL;          /* Palette set to default */
        Width = MAX_CELLS;        /* Set number of cells to def value */
        CurEnv -> Gens = 0;       /* Number of gens to print, set to def value
         which indicates that printing should continue until pgm is halted */
        CurEnv -> MutPeriod = DEF_MUT;   /* Mutation period, set to default */
        CurEnv -> Scroll = 0;    /* Whether screen should be scrolled */
        RandInit(CurEnv -> Rules,10,3);     /* Initially make rules random */
        RandInit(Cells,MAX_CELLS,3);          /* Start with random cells */
        /* Use flags from arg file to set up variables used to generate lca */
        if (!InterpretFlags(Params -> Args,Params -> NumArgs,
         &CurEnv -> Background,&CurEnv -> Palette,&Width,
         CurEnv -> Rules,&CurEnv -> Gens,&CurEnv -> MutPeriod,
         &CurEnv -> Scroll,&CurEnv -> VidMode,&CurEnv -> LineNap,
	 &CurEnv -> ScreenNap,Cells,Params))
            return 0;
        else {
            Params = Params -> Next;          /* Get next param array */
            /* Convert width (number of cells) to number of bytes they occupy
                (4 cells/byte) */
            CurEnv -> Bytes = Width >> 2;
            /* Pack cells array */
            MakePacked(Cells,CurEnv -> PackedCells,Width);
            /* If another params array to do, add another element to the
                env list */
            if (Params -> Next != NULL) {
                CurEnv -> Next = MallocT(sizeof(EnvListType));
                CurEnv = CurEnv -> Next;
             }
        }
    }
    /* Make list circular (make last element point to first */
    CurEnv -> Next = EnvList;
    return 1;
}
